"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2020 Deepkit UG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.EntityState = void 0;
const type_1 = require("@deepkit/type");
const index_1 = require("../index");
const dot_prop_1 = require("dot-prop");
const core_1 = require("@deepkit/core");
const operators_1 = require("rxjs/operators");
const rxjs_1 = require("rxjs");
class EntitySubjectStore {
    constructor() {
        this.subjects = {};
        this.consumers = {};
    }
    async forkUnsubscribed(id) {
        if (!this.consumers[id]) {
            return;
        }
        this.consumers[id].count--;
        if (this.consumers[id].count === 0 && this.subjects[id]) {
            const subject = this.subjects[id];
            delete this.subjects[id];
            await subject.unsubscribe();
        }
    }
    /**
     *  If we would return the original EntitySubject and one of the consumers unsubscribes()
     *  it would unsubscribe for ALL subscribers of that particular entity item.
     *  so we fork it. The fork can be unsubscribed without touching the origin.
     */
    createFork(id, item) {
        const originSubject = this.getOrCreateSubject(id, item);
        if (!this.consumers[id]) {
            this.consumers[id] = { count: 0 };
        }
        this.consumers[id].count++;
        const forkedSubject = new index_1.EntitySubject(originSubject.getValue(), async () => {
            await this.forkUnsubscribed(id);
        });
        const sub = originSubject.subscribe((next) => {
            try {
                forkedSubject.next(next);
            }
            catch (error) {
                if (error instanceof rxjs_1.ObjectUnsubscribedError) {
                    sub.unsubscribe();
                }
            }
        }, error => {
            forkedSubject.error(error);
        }, () => {
            forkedSubject.complete();
        });
        // originSubject.subscribe(forkedSubject);
        originSubject.delete.subscribe((v) => {
            forkedSubject.deleted = v;
            forkedSubject.delete.next(v);
        });
        const patchSub = originSubject.patches.subscribe((next) => {
            forkedSubject.patches.next(next);
        }, (error) => {
            forkedSubject.patches.error(error);
        }, () => {
            forkedSubject.patches.complete();
        });
        sub.add(() => {
            patchSub.unsubscribe();
        });
        return forkedSubject;
    }
    getEntitySubjectCount() {
        return core_1.getObjectKeysSize(this.subjects);
    }
    getForkCount(id) {
        if (this.consumers[id]) {
            return this.consumers[id].count;
        }
        return 0;
    }
    getOrCreateSubject(id, item) {
        if (!this.subjects[id]) {
            if (item) {
                this.subjects[id] = new index_1.EntitySubject(item);
            }
            else {
                throw new Error('Can not create a EntitySubject without item.');
            }
        }
        return this.subjects[id];
    }
    getItem(id) {
        if (this.subjects[id]) {
            const item = this.subjects[id].getValue();
            if (item) {
                return item;
            }
        }
        throw new Error(`Not non-undefined item for in SubjectStore for ${id}`);
    }
    removeItemAndNotifyObservers(id) {
        if (this.subjects[id]) {
            try {
                this.subjects[id].next(undefined);
                this.subjects[id].complete();
            }
            catch (error) {
                console.error('Could not next subject for', id);
            }
            delete this.subjects[id];
        }
    }
    notifyForks(id) {
        this.subjects[id].next(this.subjects[id].getValue());
    }
    notifyForksAboutPatches(id, patches) {
        this.subjects[id].patches.next(patches);
    }
    setItemAndNotifyForks(id, item) {
        if (!this.subjects[id]) {
            throw new Error(`Item not found in store for $id}`);
        }
        //by calling next on the origin EntitySubject all forks get that as well.
        this.subjects[id].next(item);
    }
    hasStoreItem(id) {
        return !!this.subjects[id];
    }
}
class EntityState {
    constructor() {
        this.deleted = new rxjs_1.Subject();
        this.items = new Map();
    }
    clear() {
        this.items.clear();
    }
    getStore(classType) {
        let store = this.items.get(classType);
        if (!store) {
            store = new EntitySubjectStore;
            this.items.set(classType, store);
        }
        return store;
    }
    handleEntityMessage(stream) {
        const classSchema = type_1.getClassSchemaByName(stream.entityName);
        const classType = classSchema.classType;
        const store = this.getStore(classType);
        if (stream.type === 'entity/update') {
            if (store.hasStoreItem(stream.id)) {
                const item = type_1.jsonSerializer.for(classType).deserialize(stream.data);
                //todo, we should not overwrite it, but modify the item in-place. This prevents bugs mis-expectations.
                store.setItemAndNotifyForks(stream.id, item);
            }
            else {
                throw new Error(`${core_1.getClassName(classType)} item not found in store for ${stream.id}. Update not possible`);
            }
        }
        if (stream.type === 'entity/patch') {
            if (store.hasStoreItem(stream.id)) {
                const toVersion = stream.version;
                const item = store.getItem(stream.id);
                //we cant do a version check like `item.version < toVersion`, since exchange issues versions always from 0 when restarted
                //so we apply all incoming patches.
                if (item) {
                    //todo rework: patch supports now $inc/$unset as well, which is not compatible with serializer.
                    const patches = type_1.jsonSerializer.for(classType).partialDeserialize(stream.patch.$set || {});
                    //it's important to not patch old versions
                    for (const [i, v] of core_1.eachPair(patches)) {
                        dot_prop_1.set(item, i, v);
                    }
                    if (stream.patch.$unset) {
                        for (const [path] of core_1.eachPair(stream.patch.$unset)) {
                            dot_prop_1.delete(item, path);
                        }
                    }
                    if (stream.patch.$inc) {
                        for (const [path, value] of core_1.eachPair(stream.patch.$inc)) {
                            const old = dot_prop_1.get(item, path) || 0;
                            dot_prop_1.set(item, path, old + value);
                        }
                    }
                    item.version = toVersion;
                    try {
                        store.notifyForksAboutPatches(stream.id, patches);
                        store.notifyForks(stream.id);
                    }
                    catch (error) {
                        console.log(`Could not notify EntitySubject #${stream.id} forks, due to ${error}`);
                    }
                }
            }
            else {
                console.debug(new Error(`${core_1.getClassName(classType)} item not found in store for ${stream.id}. Patch not possible`));
            }
        }
        if (stream.type === 'entity/remove') {
            if (store.hasStoreItem(stream.id)) {
                this.deleted.next(store.getItem(stream.id));
                store.removeItemAndNotifyObservers(stream.id);
            }
            else {
                console.debug(new Error(`${core_1.getClassName(classType)} item not found in store for ${stream.id}. Removing not possible`));
            }
        }
    }
    // public hasEntitySubject<T extends IdInterface>(classType: ClassType<T>, id: string): boolean {
    //     const store = this.getStore(classType);
    //     return store.hasStoreItem(id);
    // }
    /**
     * Creates the origin EntitySubject and returns a fork from it.
     * Origin will be removed as soon as all forks have unsubscribed.
     *
     * @param classType
     * @param jsonItem
     */
    handleEntity(classType, jsonItem) {
        const store = this.getStore(classType);
        const item = type_1.jsonSerializer.for(classType).deserialize(jsonItem);
        return store.createFork(item.id, item);
    }
    handleCollectionNext(collection, stream) {
        const classType = collection.classType;
        const store = this.getStore(classType);
        if (stream.type === 'set') {
            const setItems = [];
            for (const itemRaw of stream.items) {
                if (!collection.entitySubjects[itemRaw.id]) {
                    const item = type_1.jsonSerializer.for(classType).deserialize(itemRaw);
                    const subject = store.createFork(item.id, item);
                    setItems.push(subject.getValue());
                    collection.entitySubjects[itemRaw.id] = subject;
                    subject.pipe(operators_1.skip(1)).subscribe((i) => {
                        if (!subject.deleted) {
                            collection.deepChange.next(i);
                            //when item is removed, we get that signal before the collection gets that information. Which means we trigger loaded() twice
                            collection.seItem(i.id, i);
                            collection.loaded();
                        }
                    });
                }
                else {
                    setItems.push(collection.entitySubjects[itemRaw.id].value);
                }
            }
            collection.set(setItems);
        }
        if (stream.type === 'removeMany') {
            for (const id of stream.ids) {
                if (collection.entitySubjects[id]) {
                    collection.entitySubjects[id].unsubscribe();
                    delete collection.entitySubjects[id];
                }
            }
            collection.removeMany(stream.ids);
        }
        if (stream.type === 'sort') {
            collection.setSort(stream.ids);
        }
        if (stream.type === 'batch/start') {
            collection.batchStart();
        }
        if (stream.type === 'batch/end') {
            collection.batchEnd();
        }
        if (stream.type === 'pagination') {
            //when controller/entity-storage detected changes in those parameters, we set it without triggering an event.
            //triggering an event using setPage() etc would reload the current page.
            if (stream.event.type === 'server:change') {
                collection.pagination.setItemsPerPage(stream.event.itemsPerPage);
                collection.pagination.setPage(stream.event.page);
                collection.pagination.setTotal(stream.event.total);
                collection.pagination.setSort(stream.event.order);
                collection.pagination.setParameters(stream.event.parameters);
            }
            if (stream.event.type === 'server:apply/finished') {
                collection.pagination._applyFinished();
            }
        }
        if (stream.type === 'remove') {
            collection.remove(stream.id);
            if (collection.entitySubjects[stream.id]) {
                collection.entitySubjects[stream.id].unsubscribe();
                delete collection.entitySubjects[stream.id];
            }
        }
        if (stream.type === 'add') {
            if (!collection.entitySubjects[stream.item.id]) {
                const item = type_1.jsonSerializer.for(classType).deserialize(stream.item);
                const subject = store.createFork(item.id, item);
                collection.entitySubjects[item.id] = subject;
                collection.add(subject.getValue());
                subject.pipe(operators_1.skip(1)).subscribe((i) => {
                    if (!subject.deleted) {
                        collection.deepChange.next(i);
                        //when item is removed, we get that signal before the collection gets that information. Which means we trigger loaded() twice
                        collection.seItem(i.id, i);
                        collection.loaded();
                    }
                });
            }
        }
    }
}
exports.EntityState = EntityState;
//# sourceMappingURL=entity-state.js.map